import collections
import itertools

import numpy as np

__all__ = [
    'isconst',
    'meshshape',
    'meshndim',
    'meshgrid',
    'inv_perm',
    'unique_rows',
    'intersect_rows',
    'repeat',
]

def isconst(x, axis=None, **kwargs):
    x = np.asanyarray(x)
    
    if axis is None:
        x = x.reshape(-1)
    else:
        if isinstance(axis, int):
            axis = [axis]
        if len(axis) == 0:
            return np.ones(x.shape, dtype=bool)
        axis = sorted([d % x.ndim for d in axis])[::-1]
        for d in axis:
            x = np.moveaxis(x, d,-1)
        x = x.reshape(*x.shape[:-len(axis)],-1)
        
    if np.issubdtype(x.dtype, np.floating):
        return np.isclose(x[...,:-1], x[...,1:], **kwargs).all(axis=-1)
    return (x[...,:-1] == x[...,1:]).all(axis=-1)

def meshshape(*shapes, indexing='ij'):
    """
    The fundamental step of meshgrid, which is useful on its own as a standalone
    function. Given n shapes
    (s^1_1, ..., s^1_{D_1}), (s^2_1, ..., s^2_{D_2}), ..., (s^n_1, ..., s^n_{D_n}),
    Yields shapes
    (s^1_1, ..., s^1_{D_1}, 1, ..............., 1, ..., 1, ..............., 1),
    (1, ..............., 1, s^2_1, ..., s^2_{D_2}, ..., 1, ..............., 1),
    ...
    (1, ..............., 1, 1, ..............., 1, ..., s^n_1, ..., s^n_{D_n}),
    If indexing == 'xy', yields shapes
    (1, ..............., 1, s^1_1, ..., s^1_{D_1}, ..., 1, ..............., 1),
    (s^2_1, ..., s^2_{D_2}, 1, ..............., 1, ..., 1, ..............., 1),
    ...
    (1, ..............., 1, 1, ..............., 1, ..., s^n_1, ..., s^n_{D_n}),
    """
    if not indexing in {'ij', 'xy'}:
        raise ValueError(f"indexing must 'ij' or 'xy', but got {indexing}.")

    if indexing == 'xy':
        if len(shapes) < 2:
            raise ValueError(f"Must have at least 2 shapes if indexing == 'xy', but got {len(shapes)=}.")
        
        shapes = (shapes[1], shapes[0], *shapes[2:])
        out = list(meshshape(*shapes, indexing='ij'))
        out[1], out[0] = out[0], out[1]
        yield from out
    
    sumndim = sum(len(shape) for shape in shapes)
    cumndim = 0
    for shape in shapes:
        ndim = len(shape)
        yield (1,) * cumndim + shape + (1,) * (sumndim - cumndim - ndim)
        cumndim += ndim

def meshndim(*ndims, indexing='ij'):
    """
    Same as meshshape, but yields indices instead.
    Note that reshaping using indices is faster than reshaping using shapes in numpy,
    but the opposite is true in pytorch.
    
    f1 = lambda a, n, m: a.reshape((1,) * n  + a.shape + (1,) * m)
    f2 = lambda a, n, m: a[(None,) * n + (slice(None),) * a.ndim + (None,) * m]
    shape, n, m = (3,4,100,2,6), 5, 7
    
    a = np.random.normal(shape=shape)
    %timeit f1(a, n, m)
    489 ns ± 9.55 ns per loop (mean ± std. dev. of 7 runs, 1,000,000 loops each)
    %timeit f2(a, n, m)
    472 ns ± 7.58 ns per loop (mean ± std. dev. of 7 runs, 1,000,000 loops each)

    a = torch.randn(shape)
    %timeit f1(a, n, m)
    2.62 µs ± 16.8 ns per loop (mean ± std. dev. of 7 runs, 100,000 loops each)
    %timeit f2(a, n, m)
    10.8 µs ± 102 ns per loop (mean ± std. dev. of 7 runs, 100,000 loops each)
    """
    if not indexing in {'ij', 'xy'}:
        raise ValueError(f"indexing must 'ij' or 'xy', but got {indexing}.")

    if indexing == 'xy':
        if len(shapes) < 2:
            raise ValueError(f"Must have at least 2 shapes if indexing == 'xy', but got {len(shapes)=}.")
        
        shapes = (shapes[1], shapes[0], *shapes[2:])
        out = list(meshndim(*shapes, indexing='ij'))
        out[1], out[0] = out[0], out[1]
        yield from out
    
    sumndim = sum(ndims)
    cumndim = 0
    for ndim in ndims:
        # we don't use Ellipsis because it doesn't allow for possible additional batch dimensions
        yield (None,) * cumndim + (slice(None),) * ndim + (None,) * (sumndim - cumndim - ndim)
        cumndim += ndim

def meshgrid(*arrays, indexing='ij', axes=None, sparse=False):
    """
    The most general version of np.meshgrid. Meshes together arrays of shapes
    (*00, n0[axes[0][0]], ..., n0[axes[0][1]], *01),
    (*10, n1[axes[1][0]], ..., n1[axes[1][1]], *11),
    ..., 
    (*P0, nP[axes[P][0]], ..., nP[axes[P][1]], *P1),
    indexing must be either 'ij' or 'xy'. Defaults to 'ij' like pytorch but unlike numpy.
    If axes is None, meshes over all dimensions. axes can be None, an int,
    or a length-P list/tuple of length-2 list/tuple of ints or None, where P = len(arrays).
    Negative axis has the same semantic meaning as slicing with a negative index.
    If sparse, new dimensions are of length one instead of the broadcasted length.
    
    If indexing == 'ij', returns arrays of shapes 
    (*00, n0[axes[0][0]], ..., n0[axes[0][1]], n1[axes[1][0]], ..., n1[axes[1][1]], ..., nP[axes[P][0]], ..., nP[axes[P][1]], *01),
    (*10, n0[axes[0][0]], ..., n0[axes[0][1]], n1[axes[1][0]], ..., n1[axes[1][1]], ..., nP[axes[P][0]], ..., nP[axes[P][1]], *11),
    ...,
    (*P0, n0[axes[0][0]], ..., n0[axes[0][1]], n1[axes[1][0]], ..., n1[axes[1][1]], ..., nP[axes[P][0]], ..., nP[axes[P][1]], *P1),
    
    Otherwise, returns arrays of shapes
    (*00, n1[axes[1][0]], ..., n1[axes[1][1]], n0[axes[0][0]], ..., n0[axes[0][1]], ..., nP[axes[P][0]], ..., nP[axes[P][1]], *01),
    (*10, n1[axes[1][0]], ..., n1[axes[1][1]], n0[axes[0][0]], ..., n0[axes[0][1]], ..., nP[axes[P][0]], ..., nP[axes[P][1]], *11),
    ...,
    (*P0, n1[axes[1][0]], ..., n1[axes[1][1]], n0[axes[0][0]], ..., n0[axes[0][1]], ..., nP[axes[P][0]], ..., nP[axes[P][1]], *P1),
    
    IMPORTANT: Data is NOT copied just like pytorch, unlike numpy which copies by default.
    """
    if not indexing in {'ij', 'xy'}:
        raise ValueError(f"indexing must 'ij' or 'xy', but got {indexing}.")
    
    if axes is None or isinstance(axes, int):
        axes = [(None, axes)] * len(arrays)
    elif not (
        isinstance(axes, (list, tuple)) and len(axes) == len(arrays) and 
        all(
            (
                isinstance(axis, (list, tuple)) and len(axis) == 2 and 
                all(d is None or isinstance(d, int) for d in axis)
            )
            for axis in axes
        )
    ):
        raise TypeError(
            f"""axes must be None, an int, or a length-{len(arrays)}
            list/tuple of length-2 list/tuple of None or int, but {axes=}."""
        )

    pre_shapes, shapes, post_shapes = zip(*(
        (
            () if axis[0] is None else array.shape[:axis[0]],
            array.shape[axis[0]:axis[1]],
            () if axis[1] is None else array.shape[axis[1]:],
        )
        for array, axis in zip(arrays, axes)
    ))
    shapes = list(meshshape(*shapes, indexing=indexing))
    arrays = (
        array.reshape(pre_shape + shape + post_shape)
        for array, shape, pre_shape, post_shape in zip(arrays, shapes, pre_shapes, post_shapes)
    )
    if not sparse:
        shape = np.broadcast_shapes(*shapes)
        arrays = (
            np.broadcast_to(array, pre_shape + shape + post_shape)
            for array, pre_shape, post_shape in zip(arrays, pre_shapes, post_shapes)
        )
    
    return list(arrays)


def inv_perm(p):
    """
    Code taken from: https://stackoverflow.com/questions/11649577/how-to-invert-a-permutation-array-in-numpy
    Return an array s with which np.array_equal(arr[p][s], arr) is True.
    The array_like argument p must be some permutation of 0, 1, ..., len(p)-1.
    """
    p = np.asanyarray(p) # in case p is a tuple, etc.
    s = np.empty_like(p)
    s[p] = np.arange(p.size)
    return s

def unique_rows(arr, sorted=True, return_index=False, return_inverse=False, return_counts=False):
    """
    Code modified from https://github.com/scikit-image/scikit-image/blob/main/skimage/util/unique.py
    Much faster than np.unique(arr, axis=0)
    See https://github.com/numpy/numpy/issues/11136
    Set sorted=False for maximum speed.
    In this implementation, equal_nan=True.
    
    For speed comparison:
    N, M = 100000,100
    arr = np.random.randint(M, size=(N,2)).astype(float)
    
    %timeit unique_rows(arr, sorted=False, return_index=True, return_inverse=True)
    25.9 ms ± 386 µs per loop (mean ± std. dev. of 7 runs, 10 loops each)
    
    %timeit np.unique(arr, return_index=True, return_inverse=True, axis=0)
    85.8 ms ± 240 µs per loop (mean ± std. dev. of 7 runs, 10 loops each)
    """
    if arr.ndim != 2:
        raise ValueError("unique_rows() only makes sense for 2D arrays, "
                         f"got {arr.ndim}")
    # the view in the next line only works if the array is C-contiguous
    arr = np.ascontiguousarray(arr)
    # np.unique() finds identical items in a raveled array. To make it
    # see each row as a single item, we create a view of each row as a
    # byte string of length itemsize times number of columns in `ar`
    arr_row_view = arr.view(f"|S{arr.itemsize * arr.shape[1]}")
    out = np.unique(arr_row_view, return_index=True, return_inverse=return_inverse, return_counts=return_counts)
    out = list(out)
    out[0] = arr[out[1]]
    
    if sorted:
        idx = np.lexsort(out[0].T[::-1])
        out[0] = out[0][idx]
        if return_index:
            out[1] = out[1][idx]
        if return_inverse:
            out[2] = inv_perm(idx)[out[2]]
        if return_counts:
            i = 3 if return_inverse else 2
            out[i] = out[i][idx]
    
    if not return_index:
        out.pop(1)

    if len(out) == 1:
        return out[0]
    return out

def intersect_rows(larr, rarr, return_indices=False, **kwargs):
    """
    intersect1d but for 2D arrays. Uses the same trick as unique_rows.
    """
    if larr.ndim != 2 or rarr.ndim != 2:
        raise ValueError("intersect_rows() only makes sense for 2D arrays, "
                         f"got {larr.ndim=}, {rarr.ndim=}")
    if larr.shape[1] != rarr.shape[1]:
        raise ValueError(f"larr and rarr must both have same number of columns, but {larr.shape[1]=}, {rarr.shape[1]=}.")
        
    N_cols = larr.shape[1]
    larr, rarr = np.ascontiguousarray(larr), np.ascontiguousarray(rarr)
    larr_row_view, rarr_row_view = larr.view(f"|S{larr.itemsize * N_cols}"), rarr.view(f"|S{rarr.itemsize * N_cols}")
    
    out = np.intersect1d(larr_row_view, rarr_row_view, return_indices=True, **kwargs)
    out = list(out)
    out[0] = larr[out[1]]
    
    if not return_indices:
        return out[0]
    return out
    
def repeat(arr, repeats, chunks=None, validate=True):
    """
    Generalized np.repeat
    Copied from @MadPhysicist's solution: https://stackoverflow.com/questions/63510977/repeat-but-in-variable-sized-chunks-in-numpy
    
    Example due to @MadPhysicist in the same link:
    arr = np.array([0, 1, 2, 10, 11, 20, 21, 22, 23])
    #               >     <  >    <  >            <
    chunks = np.array([3, 2, 4])
    repeats = np.array([1, 3, 2])
    
    repeat(arr, repeats, chunks=chunks)
    >>> np.array([0, 1, 2, 10, 11, 10, 11, 10, 11, 20, 21, 22, 23, 20, 21, 22, 23])
    # repeats:    >  1  <  >         3          <  >              2             <
    """
    if chunks is None:
        return np.repeat(arr, repeats)
    
    arr, repeats, chunks = np.asanyarray(arr), np.asanyarray(repeats), np.asanyarray(chunks)

    if validate:
        if arr.ndim != 1 or repeats.ndim != 1 or chunks.ndim != 1:
            raise ValueError(f"arr, repeats, and chunks must all be 1D, but {arr.ndim=}, {repeats.ndim=} and {chunks.ndim=}.")
        if len(repeats) != len(chunks):
            raise ValueError(f"repeats and chunks must have the same length, but {len(repeats)=} and {len(chunks)=}.")
        if chunks.sum() != len(arr):
            raise ValueError(f"sum of chunks must be the length of arr, but {chunks.sum()=} and {len(arr)=}.")
        if repeats.min() < 0:
            raise ValueError(f"repeats must be non-negative, but {repeats.min()=}.")
        if repeats.min() == 0:
            raise NotImplementedError(f"Currently does not support zero repeats, but {repeats.min()=}.")

    # # deal with zero repeats (still buggy, and this is inefficient)
    # iszero = repeats == 0
    # if iszero.any():
    #     head_indices = np.concatenate([[0], chunks]).cumsum()[:-1]
    #     index = np.arange(len(arr) - chunks[iszero].sum())
    #     offsets = np.zeros_like(index)
    #     offsets[head_indices[:-1][iszero[:-1]]] -= chunks[:-1][iszero[:-1]]
    #     index -= offsets.cumsum()
    #     return repeat(arr[index], repeats[~iszero], chunks[~iszero], validate=False)

    regions = chunks * repeats
    index = np.arange(regions.sum())

    segments = np.repeat(chunks, repeats)
    resets = np.cumsum(segments[:-1])
    offsets = np.zeros_like(index)
    offsets[resets] = segments[:-1]
    offsets[np.cumsum(regions[:-1])] -= chunks[:-1]

    index -= np.cumsum(offsets)

    out = arr[index]
    
    return out



import torch
import numpy as np

import hyclib as lib

def test_pformat():
    d = {'a': torch.Tensor([[1.0,2.0],[3.0,4.0]]), 'b': {'c': np.array([[1.0,2.0],[3.0,4.0]]), 'd': 10.0}}
    expected = "'a':\n    tensor([[1., 2.],\n            [3., 4.]])\n'b':\n    'c':\n        [[1. 2.]\n         [3. 4.]]\n    'd':\n        10.0"
    assert lib.pprint.pformat(d, verbose=True) == expected


from .core import (
    pprint,
    config,
    exceptions,
    argparse,
    logging,
    warnings,
    timeit,
    clstools,
    functools,
    itertools,
    configurable,
    random,
    datetime,
)
    
import importlib
import logging as lg

logger = lg.getLogger(__name__)

modules = [
    '.np',
    '.pt',
    '.sp',
    '.io',
    '.npf',
    '.pd',
    '.plot',
]

for module in modules:
    try:
        importlib.import_module(module, package='hyclib')
    except ImportError as err:
        logger.info(f"Did not import {module=} due to ImportError {str(err)}.")
    
del importlib, modules, module, lg

cfg = config.load_package_config('hyclib')


import pathlib
import os
import pickle
import json
import logging
import collections
import contextlib

import numpy as np
import torch
from scipy import io as sio
import pandas as pd
import h5py
import mat73
from tqdm import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from .core import itertools, exceptions

logger = logging.getLogger(__name__)

ILLEGAL_KEYS = ['CLASS', 'PYTABLES_FORMAT_VERSION', 'TITLE', 'VERSION']

def loadmat(filename, simplify_cells=True, **kwargs):
    try:
        data = sio.loadmat(filename, simplify_cells=simplify_cells, **kwargs)
    except NotImplementedError as err:
        data = mat73.loadmat(filename)
    return data

class ItemProcessed(Exception):
    pass

class Counter:
    def __init__(self):
        self.count = 0

    def update(self, n=1):
        self.count += n

def desc_hdf(filename, progress=True):
    d = {}
    
    with hdf5_reader(filename) as (reader, count):
        for k, v in tqdm(reader, total=count, disable=(not progress)):
            if isinstance(v, PData):
                v_ = v.select()
                n_rows = len(v_)
                if isinstance(v_, pd.DataFrame):
                    columns = v_.columns.tolist()
                    v = f'pd.DataFrame: shape ({n_rows}, {len(columns)}), columns {columns}'
                else:
                    v = f'pd.Series: length {n_rows}'
            elif isinstance(v, h5py.Dataset):
                v = f'h5py.Dataset: shape {v.shape}, dtype {v.dtype}'
            else:
                v = str(v)
                ks = k.split('/')
                k = '/'.join(ks[:-1] + ['attrs'] + ks[-1:])
            itertools.dict_set(d, k, v, delimiter='/')
            
    return d

def it_to_hdf(filename, data, groupname='', callback=None, errors='raise', delimiter='/', progress=True):
    assert isinstance(data, collections.abc.Iterable)
    
    total = len(data) if isinstance(data, collections.abc.Sized) else None
    groupname = pathlib.Path('/') / groupname
    
    if callback is None:
        callback = lambda k, v, logger: None
        
    if errors == 'raise':
        def error_handler(err, k, v, kind):
            raise RuntimeError(f"Error when creating {kind} with key {k}") from err
    elif errors == 'log':
        def error_handler(err, k, v, kind):
            logger.error(f"Error when creating {kind} with key {k}: {str(err)}")
    else:
        raise ValueError(f"errors must be 'raise' or 'log', but {errors} provided.")
    
    with logging_redirect_tqdm():
        for k, v in tqdm(data, total=total, disable=(not progress)):
            if k in ILLEGAL_KEYS:
                raise ValueError(f'key cannot be one of {illegal_keys} since it conflicts with pandas metadata, but {k=} provided')
                
            ks = k.split(delimiter)
            k = '/'.join(ks)
            
            try:
                new_v = callback(k, v, logger)
            except ItemProcessed:
                continue
            else:
                v = v if new_v is None else new_v
                
            if (is_tensor:=isinstance(v, torch.Tensor)) or isinstance(v, np.ndarray):
                v = v.detach().cpu().numpy() if is_tensor else v
                
                with h5py.File(filename, 'a') as f:
                    group = f.require_group(str(groupname))
                    try:
                        group.create_dataset(k, data=v)
                    except Exception as err:
                        error_handler(err, k, v, 'dataset')
                    
            elif isinstance(v, pd.Series) or isinstance(v, pd.DataFrame):
                with pd.HDFStore(filename, 'a') as f:
                    try:
                        f[str(groupname / k)] = v
                    except Exception as err:
                        error_handler(err, k, v, 'pandas object')
                
            else:
                with h5py.File(filename, 'a') as f:
                    group = f.require_group(str(groupname))
                    subgroup = group.require_group('/'.join(ks[:-1])) if len(ks) > 1 else group

                    try:
                        subgroup.attrs[ks[-1]] = v
                    except Exception as err:
                        error_handler(err, k, v, 'attribute')
                        
def to_hdf(filename, data, **kwargs):
    if 'delimiter' in kwargs:
        raise TypeError("'delimiter' is not a valid keyword argument/")
        
    data = list(itertools.dict_iter(data, delimiter='/'))
    it_to_hdf(filename, data, **kwargs, delimiter='/')
                        
class PData:
    def __init__(self, hdfstore, key):
        assert isinstance(hdfstore, pd.HDFStore)
        self.hdfstore = hdfstore
        self.key = key
        
    def select(self, *args, **kwargs):
        return self.hdfstore.select(self.key, *args, **kwargs)
                        
def hdf5_generator(hdfstore, h5pyfile, groupname='', counter=None):
    """
    Not a true generator in the sense that it iterates through
    the whole file structure twice. However, this is due to the
    limitation of the h5py library.
    """
    groupname = pathlib.Path('/') / groupname
    
    pd_paths = []
    for k in hdfstore.keys():
        k = pathlib.Path(k)
        if k.is_relative_to(groupname):
            path = k.relative_to(groupname)
            pd_paths.append(path)
            if counter is not None:
                counter.update()
    
    group = h5pyfile[str(groupname)]
    hdf_paths = [pathlib.Path('')]
    def func(name, obj):
        path = pathlib.Path(name)
        if not any(path.is_relative_to(p) for p in pd_paths):
            hdf_paths.append(path)
            if counter is not None:
                if isinstance(obj, h5py.Group):
                    counter.update(len([k for k in obj.attrs.keys() if k not in ILLEGAL_KEYS]))
                else:
                    counter.update()
    group.visititems(func)
    
    def generator():
        for path in pd_paths:
            yield str(path), PData(hdfstore, str(groupname / path))

        for path in hdf_paths:
            obj = group[str(path)]
            if isinstance(obj, h5py.Group):
                for attr_name, attr in obj.attrs.items():
                    if attr_name not in ILLEGAL_KEYS:
                        yield str(path / attr_name), attr
            else:
                yield str(path), obj
            
    return generator()
    
@contextlib.contextmanager
def hdf5_reader(filename, **kwargs):
    hdfstore = pd.HDFStore(filename, 'r')
    h5pyfile = h5py.File(filename, 'r')
    counter = Counter()
    try:
        yield hdf5_generator(hdfstore, h5pyfile, counter=counter, **kwargs), counter.count
    finally:
        hdfstore.close()
        h5pyfile.close()
        
def load_hdf5_obj(obj):
    if isinstance(obj, PData):
        return obj.select()
    if isinstance(obj, h5py.Dataset):
        return obj[...]
    return obj

def from_hdf(filename, callback=None, progress=True, **kwargs):
    d = {}
    
    if callback is None:
        callback = lambda d, k, v, logger: None
        
    with hdf5_reader(filename, **kwargs) as (reader, count):
        for k, v in tqdm(reader, total=count, disable=(not progress)):
            try:
                out = callback(d, k, v, logger)
            except ItemProcessed:
                pass
            else:
                k, v = (k, v) if out is None else out
                v = load_hdf5_obj(v)
                itertools.dict_set(d, k, v, delimiter='/')
    
    return d

def save(path, data, extension=None, depth=-1, overwrite=False):
    """
    Saves data at path.
    If path has a suffix and extension is provided, then 
    the suffix must match the extension, and the data 
    is stored as a single file with the specified extension at path.
    If path has a suffix and no extension is provided, then
    the extension is inferred and the data is also stored
    as a single file.
    The depth parameter is ignored in both these case.
    If path does not have suffix, then path is the directory in which
    the data is stored. The data must be a dictionary.
    The depth parameter controls the depth of the directory.
    If depth=0, then path will become a depth 0 directory, and
    if depth=-1, then path will be a directory as deep as the data dictionary.
    """
    path = pathlib.Path(path)
    
    if extension is None:
        extension = path.suffix[1:]
    
    if extension == 'pkl':
        def save_func(filename, dat):
            with open(filename, 'wb') as f:
                pickle.dump(dat, f)
    elif extension == 'json':
        def save_func(filename, dat):
            with open(filename, 'w') as f:
                json.dump(dat, f, indent=4)
    else:
        raise NotImplementedError(f"Extension {extension} is not yet implemented")
    
    if path.suffix != '': # path is a filename
        suffix = path.suffix[1:]
        if suffix != extension:
            raise ValueError(f"path suffix must match extension if suffix is present. suffix: {suffix}, extension: {extension}.")
        if not overwrite and path.is_file():
            raise exceptions.PathAlreadyExists(f"The file {str(path)} already exists.")
        path.parent.mkdir(parents=True, exist_ok=True)
        save_func(path, data)
        
    else: # path is a directory
        for key, val in itertools.flatten_dict(data, depth=depth).items():
            filename = path / f"{'/'.join(key.split('.'))}.{extension}"
            filename.parent.mkdir(parents=True, exist_ok=True)
            if not overwrite and filename.is_file():
                raise exceptions.PathAlreadyExists(f"The file {str(filename)} already exists.")
            save_func(filename, val)
        
def save_data(path, data_dict, **kwargs):
    save(path, data_dict, 'pkl', **kwargs)
    
def save_config(path, config, **kwargs):
    save(path, config, 'json', **kwargs)
    
def load(path, extension=None):
    path = pathlib.Path(path)
    if not path.exists():
        raise exceptions.PathNotFound(f"The path {path} does not exist.")
        
    if extension is None:
        extension = path.suffix[1:]
        
    if extension == 'pkl':
        def load_func(filename):
            with open(filename, 'rb') as f:
                return pickle.load(f)
    elif extension == 'json':
        def load_func(filename):
            with open(filename, 'r') as f:
                return json.load(f)
    else:
        raise NotImplementedError(f"Extension {extension} is not yet implemented")
        
    if path.is_file():
        data = load_func(path)
    elif path.is_dir():
        data = {}
        for cur_path, dirnames, filenames in os.walk(path):
            if '.ipynb_checkpoints' not in cur_path:
                for filename in filenames:
                    filename = pathlib.Path(filename)
                    if filename.suffix == f'.{extension}':
                        cur_path_rel = pathlib.Path(cur_path).relative_to(path)
                        itertools.assign_dict(data, [*cur_path_rel.parts,filename.stem], load_func(os.path.join(cur_path,filename)))
    else:
        raise IOError(f"Path {path} is neither file nor directory")
        
    return data
        
def load_data(path):
    return load(path, 'pkl')

def load_config(path):
    return load(path, 'json')


import numpy as np
import numpy.ma as ma
from scipy import stats as spstats

__all__ = [
    'mean',
    'nanmean',
    'std',
    'nanstd',
    'var',
    'nanvar',
    'sem',
    'nansem',
    'count',
    'nancount',
    'cov',
    'nancov',
    'corrcoef',
    'nancorrcoef',
    'meanerr',
    'nanmeanerr',
    'weightedmeanerr',
    'nanweightedmeanerr',
    'bincount',
]

mean = np.mean
nanmean = np.nanmean

def std(x, ddof=1, **kwargs):
    return np.std(x, ddof=ddof, **kwargs)

def nanstd(x, ddof=1, **kwargs):
    return np.nanstd(x, ddof=ddof, **kwargs)

def var(x, ddof=1, **kwargs):
    return np.var(x, ddof=ddof, **kwargs)

def nanvar(x, ddof=1, **kwargs):
    return np.nanvar(x, ddof=ddof, **kwargs)
    
def sem(x, axis=None, ddof=1, **kwargs):
    return std(x, axis=axis, ddof=ddof, **kwargs) / count(x, axis=axis)**0.5
    # return spstats.sem(x, axis=axis, ddof=ddof)

def nansem(x, axis=None, ddof=1, **kwargs):
    return nanstd(x, axis=axis, ddof=ddof, **kwargs) / nancount(x, axis=axis)**0.5
    # return spstats.sem(x, axis=axis, ddof=ddof, nan_policy='omit')
    
def count(x, axis=None):
    return np.sum(np.ones(x.shape, dtype=int), axis=axis)

def nancount(x, axis=None):
    return np.sum(~np.isnan(x), axis=axis)
    
cov = np.cov # by default, np.cov uses Bessel correction, in contrast to the rest of its stats functions
    
def nancov(x, y=None, **kwargs):
    x = ma.masked_where(np.isnan(x), x)
    if y is not None:
        y = ma.masked_where(np.isnan(y), y)
    return ma.cov(x, y=y, **kwargs)
    
corrcoef = np.corrcoef

def nancorrcoef(x, y=None, **kwargs):
    x = ma.masked_where(np.isnan(x), x)
    if y is not None:
        y = ma.masked_where(np.isnan(y), y)
    return ma.corrcoef(x, y=y, **kwargs)

def meanerr(y, yerr, axis=None):
    y, yerr = np.array(y), np.array(yerr)
    
    assert y.shape == yerr.shape
    
    y = np.mean(y, axis=axis)
    yerr = np.sum(yerr**2, axis=axis)**0.5/count(yerr, axis=axis)
        
    return y, yerr

def nanmeanerr(y, yerr, axis=None):
    y, yerr = np.array(y), np.array(yerr)
    
    assert y.shape == yerr.shape
    
    nans = np.isnan(y) | np.isnan(yerr)
    y, yerr = y.copy(), yerr.copy()
    y[nans] = np.nan
    yerr[nans] = np.nan

    y = np.nanmean(y, axis=axis)
    yerr = np.nansum(yerr**2, axis=axis)**0.5/nancount(yerr, axis=axis)
        
    return y, yerr

def weightedmeanerr(y, yerr, axis=None):
    """
    Weighted mean and error of weighted mean
    
    Reference:
    Bevington and Robinson - Data Reduction and Error Analysis
    """
    y, yerr = np.array(y), np.array(yerr)
    
    assert y.shape == yerr.shape
    
    y = np.sum(y/yerr**2, axis=axis) / np.sum(1/yerr**2, axis=axis) # eq. 4.17 in reference
    yerr = 1/np.sum(1/yerr**2, axis=axis)**0.5 # eq 4.19 with square root
    
    return y, yerr
    
def nanweightedmeanerr(y, yerr, axis=None):
    """
    Weighted mean and error of weighted mean with np.nan values (either in mean or in error) ignored in the calculations.
    
    Reference:
    Bevington and Robinson - Data Reduction and Error Analysis
    """
    y, yerr = np.array(y), np.array(yerr)
    
    assert y.shape == yerr.shape
    
    nans = np.isnan(y) | np.isnan(yerr)
    y, yerr = y.copy(), yerr.copy()
    y[nans] = np.nan
    yerr[nans] = np.nan
    
    y = np.nansum(y/yerr**2, axis=axis) / np.nansum(1/yerr**2, axis=axis) # eq. 4.17 in reference
    yerr = 1/np.nansum(1/yerr**2, axis=axis)**0.5 # eq 4.19 with square root
    
    return y, yerr

def bincount(x, weights=None, minlength=0, nan_policy='propagate'):
    if nan_policy == 'propagate' or weights is None:
        return np.bincount(x, weights=weights, minlength=minlength)
    
    if nan_policy == 'omit':
        mask = np.isnan(weights)
        minlength = max(x.max() + 1, minlength)
        x = x[~mask]
        weights = weights[~mask]
        return np.bincount(x, weights=weights, minlength=minlength)
    
    raise ValueError(f"nan_policy must be 'propagate' or 'omit', but {nan_policy=}.")


from time import perf_counter
import logging

logger = logging.getLogger(__name__)

# copied from https://stackoverflow.com/questions/33987060/python-context-manager-that-measures-time
class timeit:
    def __init__(self, name=None, float_format='{:.4f}'):
        self.name = name
        self.float_format = float_format
    
    def __enter__(self):
        self.time = perf_counter()
        return self

    def __exit__(self, type, value, traceback):
        self.time = perf_counter() - self.time
        prefix = f'Time' if self.name is None else f'Time in block {self.name}'
        self.readout = f'{prefix}: {self.float_format.format(self.time)} seconds'
        logger.info(self.readout)


import pathlib

import hyclib as lib

def test_save_load(tmp_path):
    A = {'abc': 'bcd', 'abcde': {'cde': 'def'}}
    
    # data_path = pathlib.Path('/home/hc3190/lib/test/data')
    # lib.io.save_data(data_path, A)
    
    lib.io.save_data(tmp_path, A)
    new_A = lib.io.load_data(tmp_path)
    assert lib.itertools.flatten_dict(A) == lib.itertools.flatten_dict(new_A)
    
def test_save_load_2(tmp_path):
    A = {'abc': 'bcd', 'abcde': {'cde': 'def'}}
    
    # data_path = pathlib.Path('/home/hc3190/lib/test/data')
    # data_path = data_path / 'test_data.pkl'
    # lib.io.save_data(data_path, A)
    
    tmp_path = tmp_path / 'test_data.pkl'
    lib.io.save_data(tmp_path, A)
    new_A = lib.io.load_data(tmp_path)
    assert lib.itertools.flatten_dict(A) == lib.itertools.flatten_dict(new_A)


import random
import contextlib

@contextlib.contextmanager
def set_seed(seed):
    python_state = random.getstate()
    random.seed(seed)
    
    try:
        import numpy as np
    except ImportError:
        np_state = None
    else:
        np_state = np.random.get_state()
        np.random.seed(seed)
        
    try:
        import torch
    except ImportError:
        torch_state = None
    else:
        torch_state = torch.random.get_rng_state()
        torch.random.manual_seed(seed)
        
    try:
        yield
    finally:
        random.setstate(python_state)
        
        if np_state is not None:
            np.random.set_state(np_state)
            
        if torch_state is not None:
            torch.random.set_rng_state(torch_state)


import pickle

import hyclib as lib

class hi:
    def __call__(self, a, b, c, d, e, f):
        return f'{a=}, {b=}, {c=}, {d=}, {e=}, {f=}'
        
    def __repr__(self):
        return "HI"
    
def test_rpartial():
    rhi = lib.functools.rpartial(hi(), 1, d=4)

    assert repr(rhi) == 'rpartial(HI, 1, d=4)'
    assert rhi(2, 3, e=5, f=6) == 'a=2, b=3, c=1, d=4, e=5, f=6'

    rhi2 = pickle.loads(pickle.dumps(rhi))
    
    assert repr(rhi2) == 'rpartial(HI, 1, d=4)'
    assert rhi2(2, 3, e=5, f=6) == 'a=2, b=3, c=1, d=4, e=5, f=6'

    rrhi = lib.functools.rpartial(rhi, 2, e=5)

    assert repr(rrhi) == 'rpartial(HI, 2, 1, d=4, e=5)'
    assert rrhi(3, f=6) == 'a=3, b=2, c=1, d=4, e=5, f=6'


import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.offsetbox import AnchoredText
import statsmodels.api as sm
import numpy as np

import hyclib as lib

def subplots(rows, cols, plot_size=(6.4,4.8), keep_shape=False, **kwargs):
    fig, axes = plt.subplots(rows, cols, figsize=(plot_size[0]*cols, plot_size[1]*rows), **kwargs)
    if keep_shape:
        axes = np.array(axes)
        axes = axes.reshape((rows, cols))
    return fig, axes

def scaterr(x, y, yerr, ax=None, cap=False, **kwargs):
    if 'marker' not in kwargs:
        kwargs['marker'] = '.'
    if 'ls' not in kwargs and 'linestyle' not in kwargs:
        kwargs['ls'] = 'none'
    if 'capsize' not in kwargs and cap:
        kwargs['capsize'] = 2.0
    
    if ax is None:
        ax = plt.gca()
    
    return ax.errorbar(x.tolist(), y.tolist(), yerr=yerr.tolist(), **kwargs)

def hide_unused_axes(axes):
    for ax in axes.flat:
        if not ax.has_data():
            ax.set_axis_off()
            
def errorbar(x, y, yerr=None, plot_kwargs=None, fill_kwargs=None, ax=None):
    if ax is None:
        ax = plt.gca()
        
    if plot_kwargs is None:
        plot_kwargs = {}
    
    if fill_kwargs is None:
        fill_kwargs = {}
        
    plot_kwargs = {
        'color': 'C0'
    } | plot_kwargs
    
    fill_kwargs = {
        'alpha': 0.15,
        'facecolor': 'C0',
        'edgecolor': 'none',
    } | fill_kwargs
        
    artists = {}
    artists['line'] = ax.plot(x, y, **plot_kwargs)
    artists['fill'] = ax.fill_between(x, y - yerr, y + yerr, **fill_kwargs)
    
    return artists

def plot_ci(fit_res, alpha=0.05, xlim=None, plot_kwargs=None, fill_kwargs=None, ax=None):
    if ax is None:
        ax = plt.gca()
        
    if plot_kwargs is None:
        plot_kwargs = {}
    
    if fill_kwargs is None:
        fill_kwargs = {}
        
    default_plot_kwargs = {
        'color': 'C0',
    }
    default_plot_kwargs.update(plot_kwargs)
    plot_kwargs = default_plot_kwargs
    
    default_fill_kwargs = {
        'alpha': 0.15,
        'facecolor': 'C0',
        'edgecolor': 'none',
    }
    default_fill_kwargs.update(fill_kwargs)
    fill_kwargs = default_fill_kwargs
    
    if xlim is None:
        xlim = ax.get_xlim()
    x = np.linspace(*xlim)
    pred_res = fit_res.get_prediction(sm.add_constant(x))
    summary_frame = pred_res.summary_frame(alpha=alpha)
    
    artists = {}
    artists['line'] = ax.plot(x, summary_frame['mean'], **plot_kwargs)
    artists['fill'] = ax.fill_between(x, summary_frame['mean_ci_lower'], summary_frame['mean_ci_upper'], **fill_kwargs)
    
    return artists
    
def regplot(x, y, yerr=None, scatter_kwargs=None, ci_kwargs=None, text_kwargs=None, ax=None):
    if ax is None:
        ax = plt.gca()
        
    if scatter_kwargs is None:
        scatter_kwargs = {}
        
    if ci_kwargs is None:
        ci_kwargs = {}
    
    if text_kwargs is None:
        text_kwargs = {}
        
    default_scatter_kwargs = {
        'marker': 'o',
        'ls': 'none',
    }
    default_scatter_kwargs.update(scatter_kwargs)
    scatter_kwargs = default_scatter_kwargs
    
    default_text_kwargs = {
        'loc': 'upper right',
    }
    default_text_kwargs.update(text_kwargs)
    text_kwargs = default_text_kwargs
    
    x = np.asarray(x)
    y = np.asarray(y)
    
    if yerr is None:
        model = sm.OLS(y,sm.add_constant(x))
    else:
        yerr = np.asarray(yerr)
        model = sm.WLS(y,sm.add_constant(x),weights=1/yerr**2)
        
    fit_res = model.fit()
    
    artists = {}
    artists['scatter'] = ax.errorbar(x, y, yerr=yerr, **scatter_kwargs)
    artists['ci'] = plot_ci(fit_res, ax=ax, **ci_kwargs)
    artists['text'] = ax.add_artist(AnchoredText(
        '\n'.join([
            f'm={fit_res.params[1]:.1e}$\pm${fit_res.bse[1]:.1e}',
            f'b={fit_res.params[0]:.1e}$\pm${fit_res.bse[0]:.1e}',
            f'r={fit_res.rsquared**0.5:.2f}, p={fit_res.pvalues[1]:.1e}',
        ]),
        **text_kwargs
    ))
    
    return artists

def _lineplot(data, x, y, yerr=None, weighted=False, label=None, errstyle='fill', ax=None, **kwargs):
    default_kwargs = {
        'errorbar_kwargs': {'label': label},
        'plot_kwargs': {'label': label},
        'fill_kwargs': {'alpha': 0.25},
    }
    kwargs = {k: v | (kwargs[k] if k in kwargs else v) for k, v in default_kwargs.items()}
    
    if ax is None:
        ax = plt.gca()
    
    x_, y_, yerr_ = [], [], []
    for xi, group in data.groupby(x):
        x_.append(xi)
        yi = group[y]
        if yerr is None:
            yi, yerri = lib.np.nanmean(yi), lib.np.nansem(yi)
        else:
            yerri = group[yerr]
            if weighted:
                yi, yerri = lib.np.nanweightedmeanerr(yi, yerri)
            else:
                yi, yerri = lib.np.nanmeanerr(yi, yerri)
        y_.append(yi)
        yerr_.append(yerri)
    x_, y_, yerr_ = np.array(x_), np.array(y_), np.array(yerr_)

    if errstyle == 'bar':
        l = ax.errorbar(x_, y_, yerr=yerr_, **kwargs['errorbar_kwargs'])
    elif errstyle == 'fill':
        l1 = ax.plot(x_, y_, **kwargs['plot_kwargs'])
        l2 = ax.fill_between(x_, y_-yerr_, y_+yerr_, **kwargs['fill_kwargs'])
        l = (l1, l2)
    else:
        raise ValueError(f"errstyle must be 'fill' or 'bar', but {errstyle=} given.")
    ax.set_xlabel(x)
    ax.set_ylabel(y)
    
    return l

def lineplot(data, x, y, yerr=None, hue=None, ax=None, **kwargs):
    if ax is None:
        ax = plt.gca()
        
    if hue is not None:
        groupby = data.groupby(hue)
        ls = {}
        for i, (label, group) in enumerate(groupby):
            l = _lineplot(group, x, y, yerr=yerr, label=label, color=mpl.cm.get_cmap('Blues')((i+1)/len(groupby)), ax=ax, **kwargs)
            ls[label] = l
        ax.legend(title=hue)
        return ls
    else:
        return _lineplot(data, x, y, yerr=yerr, ax=ax, **kwargs)
    
def extract_data(obj):
    if isinstance(obj, mpl.lines.Line2D):
        x, y = obj.get_data()
        x, y = x.astype(float), y.astype(float)
        
        data = {'x': x, 'y': y}
    
    elif isinstance(obj, mpl.container.ErrorbarContainer):
        l, _, lc = obj
        x, y = l.get_data()
        x, y = x.astype(float), y.astype(float)
        
        if len(lc) != 1:
            raise NotImplementedError()
        lc = lc[0]
        yerr = np.zeros((len(x), 2))
        for i, l in enumerate(lc.get_segments()):
            yerr[i,:] = l[:,1]
            
        if np.allclose(0.5*(yerr[:,0] + yerr[:,1]), y):
            yerr = yerr[:,1] - y
            
        data = {'x': x, 'y': y, 'yerr': yerr}
    
    else:
        raise NotImplementedError(f'Object with dtype {obj.dtype} not implemented')
        
    return data

def vec(vec, origin=None, ax=None, pad=0.05, **kwargs):
    assert len(vec) == 2
    
    default_kwargs = {
        'angles': 'xy',
        'scale_units': 'xy',
        'scale': 1
    }
    default_kwargs.update(kwargs)
    
    if ax is None:
        ax = plt.gca()
        
    if origin is None:
        origin = [0,0]
        
    arrow = ax.quiver(*origin, *vec, **default_kwargs)
    
    xlim, ylim = ax.get_xlim(), ax.get_ylim()
    
    xmin = min(origin[0], origin[0]+vec[0])
    xmax = max(origin[0], origin[0]+vec[0])
    xrange = xmax - xmin
    xmin = min(xlim[0], xmin-pad*xrange)
    xmax = max(xlim[1], xmax+pad*xrange)
    
    ymin = min(origin[1], origin[1]+vec[1])
    ymax = max(origin[1], origin[1]+vec[1])
    yrange = ymax - ymin
    ymin = min(ylim[0], ymin-pad*yrange)
    ymax = max(ylim[1], ymax+pad*yrange)
    
    ax.set_xlim(xmin, xmax)
    ax.set_ylim(ymin, ymax)
    
    return arrow

def set_aspect(option, ax=None):
    if ax is None:
        ax = plt.gca()
    
    if option == 'equal_square':
        ax.set_aspect('equal')
        xlim, ylim = ax.get_xlim(), ax.get_ylim()
        xrange = xlim[1] - xlim[0]
        yrange = ylim[1] - ylim[0]
        
        if xrange < yrange:
            ratio = yrange / xrange
            xcenter = 0.5*(xlim[0] + xlim[1])
            xlim = ((xlim[0]-xcenter)*ratio+xcenter, (xlim[1]-xcenter)*ratio+xcenter)
        else:
            ratio = xrange / yrange
            ycenter = 0.5*(ylim[0] + ylim[1])
            ylim = ((ylim[0]-ycenter)*ratio+ycenter, (ylim[1]-ycenter)*ratio+ycenter)
        ax.set_xlim(xlim)
        ax.set_ylim(ylim)
        
    else:
        ax.set_aspect(option)
        
def plot_matlab_fig(filename, size=(4, 3)):
    location_map = {
        'north': 'upper center',
        'south': 'lower center',
        'east': 'right',
        'west': 'left',
        'northeast': 'upper right',
        'northwest': 'upper left',
        'southeast': 'lower right',
        'southwest': 'lower left',
        'best': 'best',
        'none': 'best',
    }
    
    data = lib.io.loadmat(filename, squeeze_me=True, struct_as_record=False, simplify_cells=False)
    matfig = data['hgS_070000']
    childs = np.atleast_1d(matfig.children)
    
    if len([c for c in childs if c.type == 'axes']) > 1:
        # Position - (left, bottom, width, height)
        positions = np.stack([c.properties.Position for c in childs if c.type == 'axes'])
        
        n_cols = len(np.unique(positions[:,0]))
        n_rows = len(np.unique(positions[:,1]))
    else:
        n_cols, n_rows = 1, 1
    
    x, y = size[0] * n_cols, size[1] * n_rows
    
    fig, axes = plt.subplots(n_rows, n_cols, figsize=(x, y), squeeze=False)
    axes = axes.reshape(-1)
    
    i = 0
    for c in childs:
        if c.type == 'axes':
            ax = axes[i]
            
            title_idx, xlabel_idx, ylabel_idx, _ = c.special - 1
            
            for j, line in enumerate(c.children):
                if line.type in ['graph2d.lineseries', 'specgraph.errorbarseries']:
                    capsize = getattr(line.properties, 'CapSize', None)
                    color = getattr(line.properties, 'Color', (0, 0, 1))
                    linestyle = getattr(line.properties, 'LineStyle', '-')
                    linewidth = getattr(line.properties, 'LineWidth', None)
                    marker = getattr(line.properties, 'Marker', 'none')
                    marker_size = getattr(line.properties, 'MarkerSize', None)
                    label = getattr(line.properties, 'DisplayName', None)
            
                    x = np.atleast_1d(line.properties.XData)
                    y = np.atleast_1d(line.properties.YData)
                    
                    if line.type == 'graph2d.lineseries':
                        ax.plot(x, y, color=color, linestyle=linestyle, linewidth=linewidth, marker=marker, ms=marker_size, label=label)
                    else:
                        lerr = np.atleast_1d(line.properties.LData)
                        uerr = np.atleast_1d(line.properties.UData)
                        ax.errorbar(x, y, yerr=(lerr, uerr), capsize=capsize, color=color, linestyle=linestyle, linewidth=linewidth, marker=marker, ms=marker_size, label=label)
                    
                elif line.type == 'text':
                    text = line.properties.String
                    interpreter = getattr(line.properties, 'Interpreter', None)
                    fontsize = getattr(line.properties, 'FontSize', None)
                    
                    if j == xlabel_idx:
                        ax.set_xlabel(text, fontsize=fontsize)
                        
                    elif j == ylabel_idx:
                        ax.set_ylabel(text, fontsize=fontsize)
                        
                    elif j == title_idx:
                        ax.set_title(text, fontsize=fontsize)
                    
            ax.grid(getattr(c.properties, 'XGrid', False))

            if hasattr(c.properties, 'XTick'):
                rotation = getattr(c.properties, 'XTickLabelRotation', None)
                ax.set_xticks(c.properties.XTick, c.properties.XTickLabel, rotation=rotation)

            if hasattr(c.properties, 'YTick'):
                rotation = getattr(c.properties, 'YTickLabelRotation', None)
                ax.set_yticks(c.properties.YTick, c.properties.YTickLabel, rotation=rotation)
                
            ax.set_xlim(c.properties.XLim)
            ax.set_ylim(c.properties.YLim)
            
            i += 1
            
        elif c.type == 'scribe.legend':
            location = c.properties.Location
            fontsize = getattr(c.properties, 'FontSize', None)
            plt.legend(loc=location_map[location])
            
    fig.tight_layout()
    
    return fig, axes.reshape(n_rows, n_cols)


from operator import index
import builtins

import torch
import numpy as np

__all__ = [
    'nanmax',
    'nanmin',
    'bincount',
]

def nanmax(t, dim=None):
    t = t.nan_to_num(nan=-torch.inf, posinf=torch.inf, neginf=-torch.inf)
    if dim is None:
        return t.max()
    return t.max(dim=dim)

def nanmin(t, dim=None):
    t = t.nan_to_num(nan=torch.inf, posinf=torch.inf, neginf=-torch.inf)
    if dim is None:
        return t.min()
    return t.min(dim=dim)

def _bincount(indices, weights=None, minlength=0):
    if weights is None or (not weights.requires_grad and indices.shape == weights.shape):
        # only need the custom implementation of bincount if we have weights and 
        # either weights has gradient or weights has a batch dimension.
        # the pytorch implementation of bincount is much faster when weights is None for very
        # big inputs since it does not need to create the weights tensor
        return torch.bincount(indices, weights=weights, minlength=minlength)
    
    if indices.is_floating_point():
        raise TypeError(f"indices must be a tensor of integer dtype, but {indices.dtype=}.")
        
    if indices.ndim != 1 or weights.ndim < 1:
        raise ValueError(f"indices must be 1D and weights must be at least 1D, but {indices.ndim=} and {weights.ndim=}.")
        
    if len(indices) != weights.shape[-1]:
        raise ValueError(f"Length of indices must be same as last dimension of weights, but {len(indices)=} and {weights.shape[-1]=}.")
    
    if indices.device != weights.device:
        raise ValueError(f"indices and weights must be on the same device, but {indices.device=} and {weights.device=}.")
    
    if len(indices) > 0 and indices.min() < 0:
        raise ValueError(f"indices must not be negative, but {indices.min()=}.")

    shape = (*weights.shape[:-1], max(indices.max().item()+1, minlength) if len(indices) > 0 else minlength)
    
    indices = indices.broadcast_to(weights.shape)
    t = torch.zeros(shape, dtype=weights.dtype, device=weights.device)
    t.scatter_add_(-1, indices, weights)
    
    return t

def bincount(x, weights=None, minlength=0, nan_policy='propagate'):
    """
    Similar to torch.bincount, but supports auto-differentiation and allows batched weights.
    Always performs bincount on the last dimension, with the leading dimensions interpreted as batch dimensions.
    
    Benchmark:
    
    N, M = 100000, 100
    a = torch.randint(M, size=(N,))
    w = torch.normal(mean=0.0, std=1.0, size=(N,))
    a_np, w_np = a.numpy(), w.numpy()

    %timeit lib.pt.bincount(a, weights=w)
    116 µs ± 377 ns per loop (mean ± std. dev. of 7 runs, 10,000 loops each)
    
    %timeit a.bincount(weights=w)
    93.5 µs ± 187 ns per loop (mean ± std. dev. of 7 runs, 10,000 loops each)
    
    %timeit np.bincount(a_np, weights=w_np)
    161 µs ± 727 ns per loop (mean ± std. dev. of 7 runs, 10,000 loops each)
    """
    if nan_policy == 'propagate' or weights is None:
        return _bincount(x, weights=weights, minlength=minlength)
    
    if nan_policy == 'omit':
        mask = weights.isnan()
        minlength = max(x.max().item() + 1, minlength)
        x = x[~mask]
        weights = weights[~mask]
        return _bincount(x, weights=weights, minlength=minlength)
    
    raise ValueError(f"nan_policy must be 'propagate' or 'omit', but {nan_policy=}.")

def _bin_edges(sample, bins=None, range=None):
    """ 
    Create edge arrays
    """
    Dlen, Ndim = sample.shape # Dlen is batch size, Ndim is number of dimensions
    device = sample.device

    edges = Ndim * [None]         # Bin edges for each dim (will be 2D array)
    dedges = Ndim * [None]        # Spacing between edges (will be 2D array)

    # Select range for each dimension
    # Used only if number of bins is given.
    if range is None:
        smin = np.atleast_1d(nanmin(sample, dim=0).values.detach().cpu().numpy())
        smax = np.atleast_1d(nanmax(sample, dim=0).values.detach().cpu().numpy())
    else:
        if len(range) != Ndim:
            raise ValueError(
                f"range given for {len(range)} dimensions; {Ndim} required")
        smin = np.empty(Ndim)
        smax = np.empty(Ndim)
        for i in builtins.range(Ndim):
            if range[i][1] < range[i][0]:
                raise ValueError(
                    "In {}range, start must be <= stop".format(
                        f"dimension {i + 1} of " if Ndim > 1 else ""))
            smin[i], smax[i] = range[i]

    # Make sure the bins have a finite width.
    for i in builtins.range(len(smin)):
        if smin[i] == smax[i]:
            smin[i] = smin[i] - .5
            smax[i] = smax[i] + .5

    # Preserve sample floating point precision in bin edges
    edges_dtype = (sample.dtype if torch.is_floating_point(sample)
                   else torch.float)

    # Create edge arrays
    for i in builtins.range(Ndim):
        bins_i = torch.as_tensor(bins[i])
        if bins_i.ndim == 0: # scalar, interpret as number of bins
            edges[i] = torch.linspace(smin[i], smax[i], bins_i.item() + 1,
                                      dtype=edges_dtype).to(device) # compute on cpu first since MPS seems to have precision issue
        else:
            edges[i] = torch.as_tensor(bins_i, dtype=edges_dtype).to(device)
        dedges[i] = torch.diff(edges[i])

    return edges, dedges


def _bin_numbers(sample, edges, dedges):
    """Compute the bin number each sample falls into, in each dimension
    """
    Dlen, Ndim = sample.shape

    sampBin = [
        torch.bucketize(sample[:, i], edges[i], right=True) # `right` argument has opposite meaning in torch.bucketize vs np.digitize
        for i in range(Ndim)
    ]

    # Using `bucketize`, values that fall on an edge are put in the right bin.
    # For the rightmost bin, we want values equal to the right
    # edge to be counted in the last bin, and not as an outlier.
    for i in range(Ndim):
        # Find the rounding precision
        dedges_min = dedges[i].min()
        if dedges_min == 0:
            raise ValueError('The smallest edge difference is numerically 0.')
        if dedges_min < 0:
            raise ValueError('Edges must be monotonically increasing.')
            
        decimal = -torch.log10(dedges_min).long().item() + 6
        # Find which points are on the rightmost edge.
        on_edge = ((sample[:, i] >= edges[i][-1]) &
                   (torch.round(sample[:, i], decimals=decimal) ==
                    torch.round(edges[i][-1], decimals=decimal))).nonzero().squeeze()
        # Shift these points one bin to the left.
        sampBin[i][on_edge] -= 1

    return torch.stack(sampBin)

def bin_dd(sample, bins=10, range=None, nan_policy='raise'):
    """
    Bins N-dimensional data. Arguments have the same meaning as in scipy.stats.binned_statistic_dd,
    except that here expand_binnumbers=True by default, and that if expand_binnumbers=True and N = 1,
    then binnumbers is 2D instead of 1D as in scipy.stats.binned_statistic_dd.
    nan_policy can be 'raise' or 'omit'. 
    If nan_policy='raise', then ValueError is raised if sample contains any NaNs (this is slightly different
    from the default behavior of scipy.stats.binned_statistic_dd).
    If nan_policy='omit', NaNs are sorted into the rightmost bin.
    """
    
    if nan_policy not in ['raise', 'omit']:
        raise ValueError(f"nan_policy must be 'raise' or 'omit', but {nan_policy=}.")
    
    try:
        bins = index(bins)
    except TypeError:
        # bins is not an integer
        pass
    # If bins was an integer-like object, now it is an actual Python int.
    
    # `Ndim` is the number of dimensions (e.g. `2` for `binned_statistic_2d`)
    # `Dlen` is the length of elements along each dimension.
    # This code is based on np.histogramdd
    try:
        # `sample` is an ND-array.
        Dlen, Ndim = sample.shape
    except (AttributeError, ValueError):
        # `sample` is a sequence of 1D arrays.
        sample = torch.stack(sample).t()
        Dlen, Ndim = sample.shape

    # NOTE: for _bin_edges(), see e.g. gh-11365
    if nan_policy == 'raise' and not torch.isfinite(sample).all():
        raise ValueError('%r contains non-finite values.' % (sample,))
        
    try:
        M = len(bins)
        if M != Ndim:
            raise AttributeError('The dimension of bins must be equal '
                                 'to the dimension of the sample x.')
    except TypeError:
        bins = Ndim * [bins]
        
    edges, dedges = _bin_edges(sample, bins, range)
    binnumbers = _bin_numbers(sample, edges, dedges)
        
    centers = []
    for e in edges:
        c = torch.empty(len(e)+1, dtype=e.dtype, device=e.device)
        c[0] = torch.nan
        c[-1] = torch.nan
        c[1:-1] = 0.5 * (e[1:] + e[:-1])
        centers.append(c)
    
    return binnumbers, centers, edges

def bin(sample, bins=10, range=None, **kwargs):
    try:
        N = len(bins)
    except TypeError:
        N = 1

    if N != 1:
        bins = [torch.as_tensor(bins, dtype=torch.float)]

    if range is not None:
        if len(range) == 2:
            range = [range]
            
    binnumbers, centers, edges = bin_dd([sample], bins=bins, range=range, **kwargs)
    return binnumbers[0], centers[0], edges[0]